/*
 * Copyright (c) 2010 Rozhuk Ivan <rozhuk.im@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */




#define WINVER       0x0501
#define _WIN32_WINNT 0x0501


#include "SocketFunctions.h"
#include <windows.h>
#include <Shellapi.h>
//#include <strsafe.h>
#ifdef NDEBUG
	#include "memfuncs.h"
#endif //_DEBUG

#include "CRCFileCheck.h"
#include "ReportError.h"
#include "DebugFunctions.h"
#include "RawDataToPrinter.h"
#include "StrToNum.h"
#include "CalcChecksum.h"


#pragma comment(lib, "Shell32.lib")



#define _CRT_NOFORCE_MANIFEST
#pragma comment(linker,"/manifestdependency:\"type='win32' name='Microsoft.Windows.Common-Controls' version='6.0.0.0' processorArchitecture='*' publicKeyToken='6595b64144ccf1df' language='*'\"")



#ifdef NDEBUG
#ifdef  _WIN64
#pragma comment(lib,"runtmchk.lib")
//#pragma check_stack(off)
#else
extern "C" __declspec(naked) void __cdecl _chkstk()
{
	#define _PAGESIZE_ 4096

	__asm{
			push    ecx

	; Calculate new TOS.

			lea     ecx, [esp] + 8 - 4      ; TOS before entering function + size for ret value
			sub     ecx, eax                ; new TOS

	; Handle allocation size that results in wraparound.
	; Wraparound will result in StackOverflow exception.

			sbb     eax, eax                ; 0 if CF==0, ~0 if CF==1
			not     eax                     ; ~0 if TOS did not wrapped around, 0 otherwise
			and     ecx, eax                ; set to 0 if wraparound

			mov     eax, esp                ; current TOS
			and     eax, not ( _PAGESIZE_ - 1) ; Round down to current page boundary

	cs10:
			cmp     ecx, eax                ; Is new TOS
			jb      short cs20              ; in probed page?
			mov     eax, ecx                ; yes.
			pop     ecx
			xchg    esp, eax                ; update esp
			mov     eax, dword ptr [eax]    ; get return address
			mov     dword ptr [esp], eax    ; and put it at new TOS
			ret

	; Find next lower page and probe
	cs20:
			sub     eax, _PAGESIZE_         ; decrease by PAGESIZE
			test    dword ptr [eax],eax     ; probe page.
			jmp     short cs10
	}
}


#endif
#endif //_DEBUG



struct PROG_SETTINGS
{
	LPWSTR		pszFWFileName;
	LPWSTR		pszPrinterName;
	LPWSTR		pszSaveFileName;
	DWORD		dwAuto;
	int			iRemonteAddrLength;
	sockaddr_in	siRemonteAddr;
};


PROG_SETTINGS gpsProgSettings;


#define MEMALLOC(Size)				HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,(Size+sizeof(SIZE_T)))
#define MEMREALLOC(Mem,Size)		HeapReAlloc(GetProcessHeap(),(HEAP_ZERO_MEMORY),(LPVOID)Mem,(Size+sizeof(SIZE_T)))
#define MEMFREE(Mem)				if (Mem) {HeapFree(GetProcessHeap(),0,(LPVOID)Mem);Mem=NULL;}

void		Usage();
int			GetOptions(int argc,LPWSTR *argv);






DWORD HPFirmwareAddHeader(LPBYTE pFirmwareData,DWORD dwFirmwareDataSize,LPBYTE pFirmwarePatched,DWORD dwFirmwarePatchedBuffSize,DWORD *pdwFirmwarePatchedSize)
{
	DWORD dwRetErrorCode;

	if (pFirmwareData && dwFirmwareDataSize>4 && pFirmwarePatched && dwFirmwarePatchedBuffSize>(dwFirmwareDataSize+64))
	{
		DWORD dwFirmwarePatchedSize;
		unsigned char btELFFillHDR[17]=		{0xbe, 0xef, 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',/* size > */0, 0, 0, 0/* < size */, 0, 0, 0};
		unsigned char btELFSecHDR[12]=		{0xc0, 0xde, 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 0, 0};
		unsigned char btPLJFirstHDR[34]=	{0x1b, '%', '-', '1', '2', '3', '4', '5', 'X', '@', 'P', 'J', 'L', ' ', 'E', 'N', 'T', 'E', 'R', ' ', 'L', 'A', 'N', 'G', 'U', 'A', 'G', 'E', '=', 'A', 'C', 'L', 0x0d, 0x0a};
		//unsigned char btPLJSecondHDR[8]=	{0x00, 0xac, 0xc0, 0xde,/* size > */0, 0, 0, 0/* < size */};
		unsigned char btPLJSecondHDR[4]=	{0x00, 0xac, 0xc0, 0xde};
		unsigned char btPLJEndHDR[9]=		{0x1b, '%', '-', '1', '2', '3', '4', '5', 'X'};

		const DWORD PLJ_FIRST_HDR_SIZE=		sizeof(btPLJFirstHDR);
		const DWORD PLJ_SECOND_HDR_SIZE=	8;
		const DWORD PLJ_END_HDR_SIZE=		sizeof(btPLJEndHDR);



		if ((*((DWORDLONG*)pFirmwareData))==(*((DWORDLONG*)btELFFillHDR)))
		{// no need add header, elf download header found
			memmove(pFirmwarePatched,pFirmwareData,dwFirmwareDataSize);
			dwFirmwarePatchedSize=dwFirmwareDataSize;
			dwRetErrorCode=NO_ERROR;
		}else
		if ((*((DWORDLONG*)pFirmwareData))==(*((DWORDLONG*)btPLJFirstHDR)))
		{// no need add header, plj download header found
			memmove(pFirmwarePatched,pFirmwareData,dwFirmwareDataSize);
			dwFirmwarePatchedSize=dwFirmwareDataSize;
			dwRetErrorCode=NO_ERROR;
		}else
		if ((*((DWORD*)pFirmwareData))==(*((DWORD*)"\177ELF")))
		{// elf
			(*((DWORD*)&btELFFillHDR[10]))=HTONL((dwFirmwareDataSize+sizeof(btELFSecHDR)+4));// firmware size + section header + Chksum

			memmove((pFirmwarePatched+(sizeof(btELFFillHDR)+sizeof(btELFSecHDR))),pFirmwareData,dwFirmwareDataSize);// first move firmware data - we dont erase original data if dst==src (pFirmwarePatched==pFirmwareData)
			memmove(pFirmwarePatched,btELFFillHDR,sizeof(btELFFillHDR));
			memmove((pFirmwarePatched+sizeof(btELFFillHDR)),btELFSecHDR,sizeof(btELFSecHDR));// Create and write the section header

			(*((WORD*)(pFirmwarePatched+(sizeof(btELFFillHDR)+sizeof(btELFSecHDR)+dwFirmwareDataSize))))=0xFFFF;
			(*((WORD*)(pFirmwarePatched+(sizeof(btELFFillHDR)+sizeof(btELFSecHDR)+dwFirmwareDataSize+sizeof(WORD)))))=FastCalcChecksum((pFirmwarePatched+sizeof(btELFFillHDR)),(sizeof(btELFSecHDR)+dwFirmwareDataSize));// Chksum of btELFSecHDR = const

			dwFirmwarePatchedSize=(sizeof(btELFFillHDR)+sizeof(btELFSecHDR)+dwFirmwareDataSize+sizeof(DWORD));
			dwRetErrorCode=NO_ERROR;
		}else
		if ((*((WORD*)pFirmwareData))==(*((WORD*)"20")))
		{// plj
			// first: move firmware data - we dont erase original data if dst==src (pFirmwarePatched==pFirmwareData)
			memmove((pFirmwarePatched+(PLJ_FIRST_HDR_SIZE+PLJ_SECOND_HDR_SIZE)),pFirmwareData,dwFirmwareDataSize);

			// set first header
			memmove(pFirmwarePatched,btPLJFirstHDR,PLJ_FIRST_HDR_SIZE);

			// set second header
			//memmove((pFirmwarePatched+PLJ_FIRST_HDR_SIZE),btPLJSecondHDR,PLJ_SECOND_HDR_SIZE);
			(*((DWORD*)(pFirmwarePatched+PLJ_FIRST_HDR_SIZE)))=(*((DWORD*)btPLJSecondHDR));// optimized mem copy :)
			(*((DWORD*)(pFirmwarePatched+PLJ_FIRST_HDR_SIZE+sizeof(DWORD))))=HTONL((dwFirmwareDataSize-8));// firmware size - ???

			// set last header
			memmove((pFirmwarePatched+(PLJ_FIRST_HDR_SIZE+PLJ_SECOND_HDR_SIZE+dwFirmwareDataSize)),btPLJEndHDR,PLJ_END_HDR_SIZE);

			dwFirmwarePatchedSize=(PLJ_FIRST_HDR_SIZE+PLJ_SECOND_HDR_SIZE+dwFirmwareDataSize+PLJ_END_HDR_SIZE);
			dwRetErrorCode=NO_ERROR;
		}else{// unknown
			DebugBreak();
			dwRetErrorCode=ERROR_INVALID_DATA;
		}

		if (pdwFirmwarePatchedSize) (*pdwFirmwarePatchedSize)=dwFirmwarePatchedSize;
	}else{
		dwRetErrorCode=ERROR_INVALID_PARAMETER;
	}
return(0);
}
 




#ifdef _DEBUG
	#ifdef UNICODE
		int APIENTRY wWinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,LPWSTR lpCmdLine,int nCmdShow)
	#else
		int APIENTRY WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,LPSTR lpCmdLine,int nCmdShow)
	#endif
#else
	void WinMainEx(void)
#endif //_DEBUG
{
	#ifdef NDEBUG
		HINSTANCE	hInstance		=GetModuleHandle(NULL);
		HINSTANCE	hPrevInstance	=NULL;
		LPWSTR		lpCmdLine		=GetCommandLine();
		int			nCmdShow		=0;

	#endif //_DEBUG

	DWORD dwRetErrorCode;


	//Usage();

	dwRetErrorCode=CRCFileCheck();
	if (dwRetErrorCode==NO_ERROR)
	{//     
		dwRetErrorCode=SocketsInitialize();
		if (dwRetErrorCode==NO_ERROR)
		{
			if (lpCmdLine && lstrlen(lpCmdLine))
			{
				int argc;
				LPWSTR *argv;

				//MessageBox(NULL,lpCmdLine,L"lpCmdLine",(MB_OK|MB_ICONINFORMATION));
				argv=CommandLineToArgvW(lpCmdLine,&argc);
				if (argv)
				{
				#ifdef _DEBUG
					for(int i=0;i<argc;i++)
					{
						DebugPrintCRLF(argv[i]);
					}
				#endif //_DEBUG

					memset(&gpsProgSettings,0,sizeof(gpsProgSettings));
					gpsProgSettings.dwAuto=1;
					if (GetOptions(argc,argv)==0)
					{
						if (gpsProgSettings.pszFWFileName)
						{
							HANDLE hFile;
							DWORD dwFileSize,dwFirmwareDataSize;
							LPBYTE pFirmwareData;

							hFile=CreateFile(gpsProgSettings.pszFWFileName,FILE_READ_DATA,FILE_SHARE_READ,NULL,OPEN_EXISTING,(FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN),NULL);
							if (hFile!=INVALID_HANDLE_VALUE && (dwFileSize=GetFileSize(hFile,NULL))!=INVALID_FILE_SIZE)
							{
								pFirmwareData=(LPBYTE)MEMALLOC((dwFileSize+2048));
								if (pFirmwareData)
								{
									dwFirmwareDataSize=0;
									if (ReadFile(hFile,pFirmwareData,(dwFileSize+2048),&dwFirmwareDataSize,NULL) && dwFileSize==dwFirmwareDataSize)
									{

										if (gpsProgSettings.dwAuto)
										{
											dwRetErrorCode=HPFirmwareAddHeader(pFirmwareData,dwFirmwareDataSize,pFirmwareData,(dwFileSize+2048),&dwFirmwareDataSize);
										}else{
											dwRetErrorCode=NO_ERROR;
										}


										if (dwRetErrorCode==NO_ERROR)
										if (gpsProgSettings.pszSaveFileName)
										{// saving modified firmware
											HANDLE hSaveFile;
											DWORD dwNumberOfBytesWritten;

											dwRetErrorCode=NO_ERROR;
											hSaveFile=CreateFile(gpsProgSettings.pszSaveFileName,(FILE_APPEND_DATA|FILE_WRITE_DATA),FILE_SHARE_READ,NULL,CREATE_ALWAYS,(FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN),NULL);
											if (hSaveFile!=INVALID_HANDLE_VALUE)
											{
												dwNumberOfBytesWritten=0;
												if (WriteFile(hSaveFile,pFirmwareData,dwFirmwareDataSize,&dwNumberOfBytesWritten,NULL)==0 || dwFirmwareDataSize!=dwNumberOfBytesWritten)
												{// write error
													dwRetErrorCode=GetLastError();
												}

												CloseHandle(hSaveFile);
											}else{// open/create file fail
												dwRetErrorCode=GetLastError();
											}
											if (dwRetErrorCode!=NO_ERROR) ReportError(dwRetErrorCode);
											dwRetErrorCode=NO_ERROR;
										}


										if (dwRetErrorCode==NO_ERROR)
										if (gpsProgSettings.pszPrinterName)
										{// send FW trouth spoller = print raw data
											dwRetErrorCode=RawDataToPrinter(gpsProgSettings.pszPrinterName,L"PrinterFirmware",pFirmwareData,dwFirmwareDataSize);
											if (dwRetErrorCode!=NO_ERROR) ReportError(dwRetErrorCode);
											dwRetErrorCode=NO_ERROR;
										}

										if (dwRetErrorCode==NO_ERROR)
										if (gpsProgSettings.iRemonteAddrLength)
										{// send FW by TCP RAW
											int iSended,iSendedTotal;
											SOCKET skt;

											skt=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
											if (skt!=INVALID_SOCKET)
											{
												if (connect(skt,(SOCKADDR*)&gpsProgSettings.siRemonteAddr,gpsProgSettings.iRemonteAddrLength)!=SOCKET_ERROR)
												{
													iSendedTotal=0;
													do{
														DebugBreakIf(iSendedTotal);
														iSended=send(skt,(char*)(pFirmwareData+iSendedTotal),(dwFirmwareDataSize-iSendedTotal),0);
														if (iSended==SOCKET_ERROR)
														{
															dwRetErrorCode=WSAGetLastError();
															break;
														}else{
															iSendedTotal+=iSended;
														}
													}while(dwFirmwareDataSize>((DWORD)iSendedTotal));
												}else{// connect failed
													dwRetErrorCode=WSAGetLastError();
												}

												shutdown(skt,SD_SEND);
												closesocket(skt);

												if (dwRetErrorCode!=NO_ERROR) ReportError(dwRetErrorCode);
												dwRetErrorCode=NO_ERROR;
											}else{//socket creation failed
												dwRetErrorCode=WSAGetLastError();
											}
										}


									}else{// read error
										dwRetErrorCode=GetLastError();
									}

									MEMFREE(pFirmwareData);
								}else{// buff allocation error
									dwRetErrorCode=GetLastError();
								}

								CloseHandle(hFile);
							}else{// file open failed
								dwRetErrorCode=GetLastError();
							}
						}else{// firmware file not specified
							Usage();
						}
					}else{// options parce failed, print USAGE
						Usage();
					}
					LocalFree(argv);
				}
			}else{//no command line
				Usage();
			}

			WSACleanup();
		}
	}
	if (dwRetErrorCode!=NO_ERROR) ReportError(dwRetErrorCode);
	

#ifdef _DEBUG
	return(dwRetErrorCode);
#else
	ExitProcess(dwRetErrorCode);
#endif //_DEBUG
}




void Usage()
{
	MessageBox(NULL, \
		L"Usage: FWUploader -f FirmwareFileName [-a AutoAddHeaders] [-s NewFirmwareFileName] [-p PrinterName] [-t IPAddress:Port]\r\n \
\r\n \
by Rozhuk Ivan 2010\r\n rozhuk.im@gmail.com\r\n http://www.rozhuk.org.ru/\r\n \
\r\n \
-a	1 - default, if on program will check commands/header and add if needed, 0 - no check, no add\r\n \
-f	firmware file name, use quotes if name contain spaces\r\n \
-p	printer name, use quotes if name contain spaces\r\n \
-t	ip address and port for TCP RAW print\r\n \
-s	save file name for uploaded firmware, will contain modified firmware with commands/header\r\n \
\r\n\r\n \
Examples\r\n \
Upload by TCP RAW:\r\n \
FWUploader -f \"c:\\Firmwares for printer\\sihp1018.dl\" -t 192.168.0.10:9100\r\n \
\r\n \
Upload trouth spoller on connected printer:\r\n \
FWUploader -f c:\\sihp1018.dl -p \"HP LaserJet 1018\"\r\n \
\r\n \
Add HP download header/trailer to an ARM ELF binary and save to file:\r\n \
FWUploader -f c:\\hp1018.img -s c:\\sihp1018.dl\r\n \
\r\n \
Sent to printer trouth SMB:\r\n \
FWUploader -f c:\\hp1018.img -s \\\\172.16.0.10\\lp1\r\n \
\r\n \
						",L"Usage",(MB_OK|MB_ICONINFORMATION));

}



int GetOptions(int argc,LPWSTR *argv)
{
	int iRet=0;
	
	if (argc>1)
	{
		LPWSTR pCur;

		iRet=0;
		for (int i=(argc&1);i<argc;i++)
		{
			pCur=argv[i];

			if (*pCur++!=(*((WCHAR*)L"-")))
			{
				//MessageBox(NULL,L"Usage: ",L"Usage",(MB_OK|MB_ICONINFORMATION));
				//fprintf(stderr,"invalid option: \"%s\"\n", argv[i]);
				iRet=1;
				break;
			}

			while(*pCur && iRet==0)
			{
				if (*pCur==(*((WCHAR*)L"?")) || *pCur==(*((WCHAR*)L"h")) || *pCur==(*((WCHAR*)L"H")))
				{
					iRet=1;
				}else
				if (*pCur==(*((WCHAR*)L"a")) || *pCur==(*((WCHAR*)L"A")))
				{
					pCur++;
					if (*pCur)
					{
						gpsProgSettings.dwAuto=StrToUNum32((LPCSTR)pCur,lstrlenW(pCur));
					}else
					if (argv[++i])
					{
						gpsProgSettings.dwAuto=StrToUNum32((LPCSTR)argv[i],lstrlenW(argv[i]));
					}else{
						//fprintf(stderr,"option \"-p\" requires pid file name\n");
						iRet=1;
					}
				}else
				if (*pCur==(*((WCHAR*)L"f")) || *pCur==(*((WCHAR*)L"F")))
				{
					pCur++;
					if (*pCur)
					{
						gpsProgSettings.pszFWFileName=pCur;
					}else
					if (argv[++i])
					{
						gpsProgSettings.pszFWFileName=argv[i];
					}else{
						//fprintf(stderr,"option \"-p\" requires pid file name\n");
						iRet=1;
					}
				}else
				if (*pCur==(*((WCHAR*)L"p")) || *pCur==(*((WCHAR*)L"P")))
				{
					pCur++;
					if (*pCur)
					{
						gpsProgSettings.pszPrinterName=pCur;
					}else
					if (argv[++i])
					{
						gpsProgSettings.pszPrinterName=argv[i];
					}else{
						//fprintf(stderr,"option \"-p\" requires pid file name\n");
						iRet=1;
					}
				}else
				if (*pCur==(*((WCHAR*)L"s")) || *pCur==(*((WCHAR*)L"S")))
				{
					pCur++;
					if (*pCur)
					{
						gpsProgSettings.pszSaveFileName=pCur;
					}else
					if (argv[++i])
					{
						gpsProgSettings.pszSaveFileName=argv[i];
					}else{
						//fprintf(stderr,"option \"-p\" requires pid file name\n");
						iRet=1;
					}
				}else
				if (*pCur==(*((WCHAR*)L"t")) || *pCur==(*((WCHAR*)L"T")))
				{
					LPWSTR pRemonteAddrPort;

					pCur++;
					if (*pCur)
					{
						pRemonteAddrPort=pCur;
					}else
					if (argv[++i])
					{
						pRemonteAddrPort=argv[i];
					}else{
						//fprintf(stderr,"option \"-b\" requires bind port number\n");
						pRemonteAddrPort=NULL;
						iRet=1;
					}

					if (pRemonteAddrPort)
					{
						gpsProgSettings.iRemonteAddrLength=sizeof(sockaddr_in);
						if (WSAStringToAddressW(pRemonteAddrPort,AF_INET,NULL,(LPSOCKADDR)&gpsProgSettings.siRemonteAddr,&gpsProgSettings.iRemonteAddrLength)==SOCKET_ERROR)
						{
							iRet=1;
						}
					}
				}else{
					//fprintf(stderr,"invalid option: \"%c\"\n",*(pCur-1));
					iRet=1;
				}
			}
		}
	}else{
		iRet=1;
	}
return(iRet);
}




